! ------------------------------------------------------------------------
! 'SOLve SIngular IVPs' (SOLSIIVP)
! by Peter W. KOFLER,
! Institute of Applied Mathematics and Numerical Analysis,
! University of Technology Vienna, Vienna, Austria.
! ------------------------------------------------------------------------

MODULE types
! Define data types for SOLSIIVP.
   IMPLICIT NONE
   INTEGER, PARAMETER :: dp = KIND(1.0D0), & ! working precision real
                         i4b = SELECTED_INT_KIND(9)
END MODULE types

MODULE solsiivp
! --- Function and Algorithm ---
! This module provides an initial value problem solver for singular
! problems of ode. Near the singularity an equal spaced grid with global
! error estimation (Zadunaisky's technique) is used. Then a Runge-Kutta
! pair for local error estimation is used to complete the integration.
! --- Bibliography ---
! M.Metcalf, J.Reid,
! 'Fortran 90 Explained', Oxford University Press (1990),
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996).
! ------------------------------------------------------------------------
! Constants
! ------------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   REAL(KIND = dp), PARAMETER :: safety_factor_zadu = 0.9_dp
   REAL(KIND = dp), PARAMETER :: safety_factor_pair = 0.9_dp
!        safety_factor_ ... safety factor for new step sizes
   INTEGER(KIND = i4b), PARAMETER :: max_newton_steps = 25
!        max_newton_steps ... maximum of steps for Newton-iteration used
!                             in implicit methods
   INTEGER(KIND = i4b), PARAMETER :: switch_count_2 = 15
   INTEGER(KIND = i4b), PARAMETER :: switch_count_4 = 15
!        switch_count_ ... number of steps to switch to local error
!                          control for method 2 and 4 (in ivp_switch)
   LOGICAL, PARAMETER :: log_file = .true. ! data-output to files
   PRIVATE
   PUBLIC :: ivp_sing
! ------------------------------------------------------------------------
! Subroutines
! ------------------------------------------------------------------------
CONTAINS

SUBROUTINE ivp_sing(fcn, a, y, b, tol, ifail, dfcn, linear, method, &
                    switch, h_start, h_min, h_max, global_tol, &
                    local_tol, Newton_tol)
! Version 1.01, last modified 24-02-1998.
! --- Function and Algorithm ---
! Calculate the starting step size and switch-point. Then solve the
! singular ivp y'(t) = fcn(t, y) with equal spaced grid and global error
! estimation (Zadunaisky's technique) from a to the switch point. Continue
! to b with variable step size using an embedded pair to check the local
! error.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t,y)      ! right side of the ivp, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
      FUNCTION dfcn(t,y)     ! Jacobi-matrix of fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: dfcn(SIZE(y), SIZE(y))
      END FUNCTION dfcn
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: b, tol
!        b ... right bound of interval to perform integration
!        tol ... tolerance for integration
   REAL(KIND = dp), INTENT(in out) :: a, y(:)
!        a ... (in) left bound, (out) end position of integration
!        y ... (in) initial value y(a), (out) approximation for y(b)
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!                   0 = everything is oK
!                  -1 = illegal interval range
!                  -2 = illegal smallest step size
!                  -3 = illegal starting step size
!                  -4 = illegal tolerance
!                  -5 = illegal Newton-tolerance
!                  -6 = illegal method
!                  -7 = illegal maximum step size
!                  -8 = illegal switch point
!                   1 = could not complete integration due to h < h_min
!                   2 = Newton-iteration needed too many stNewton_tol
!                   3 = global error did not decrease during iterations
   LOGICAL, OPTIONAL, INTENT(in) :: linear
!        linear ... initial value problem is linear, default = non-linear
   INTEGER(KIND = i4b), OPTIONAL, INTENT(in) :: method
!        method ... to choose a particular method out of the following
!                   2 = at least order 2 for all integrations
!                   4 = at least order 4 (default)
   REAL(KIND = dp), OPTIONAL, INTENT(in) :: h_min, h_max, Newton_tol
!        h_min ... smallest allowed step size, default = 0.0
!        h_max ... maximum allowed step size, default = b-a
!        Newton_tol ... tolerance of Newton-iteration for implicit method,
!                  default = used global or local tolerance/10.0
   REAL(KIND = dp), OPTIONAL, INTENT(in out) :: switch, h_start, &
                                                global_tol, local_tol
!        switch ... (in) switch point, or -1.0 for auto detect,
!                   (out) used switch point, default = -1.0
!        h_start ... (in) starting step size, or -1.0 for auto detect,
!                    (out) used starting step size, default = -1.0
!        global_tol ... (in) global tolerance for Zadunaisky's method,
!                       (out) estimation of global error at switch point
!        local_tol ... (in) local tolerance for Runge-Kutta pair,
!                      (out) estimation of local error at x=b
!                  global_tol and local_tol override tol
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   LOGICAL :: opt_lin
   INTEGER(KIND = i4b) :: opt_meth
   REAL(KIND = dp) :: opt_hmin, opt_hmax, opt_hstart, opt_switch, &
                      opt_New_tol, opt_glo_tol, opt_loc_tol
!        opt_ ... copy of optional intent-in arguments to work with
! ------------------------------------------------------------------------
! Called software components
! ------------------------------------------------------------------------
!        direct ... ivp_starth, ivp_switch, ivp_zadunaisky, ivp_rkpair
!        indirect ... ludcmp, lubksb
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
! check input and set up variables
   ifail = -1                ! check interval, allow only 0 <= a < b
   IF (a < 0.0_dp) RETURN
   IF (b <= a) RETURN
   ifail = -2                ! check smallest step size (in)
   opt_hmin = 0.0_dp         ! default for h_min is 0.0
   IF (PRESENT(h_min)) opt_hmin = h_min
   IF (opt_hmin < 0.0_dp) RETURN
   IF (opt_hmin > b-a) RETURN
   ifail = -7                ! check maximum step size (in) for RK pair
   opt_hmax = b-a            ! default h_max = b-a
   IF (PRESENT(h_max)) opt_hmax = h_max
   IF (opt_hmax <= 0.0_dp) RETURN
   IF (opt_hmax > b-a) opt_hmax = b-a
   IF (opt_hmax < opt_hmin) RETURN
   ifail = -3                ! check starting step size (in out)
   opt_hstart = -1.0_dp      ! default h_start is auto detect
   IF (PRESENT(h_start)) opt_hstart = h_start
   IF (opt_hstart >= 0.0_dp) THEN ! check user input
      IF (opt_hstart < opt_hmin) RETURN
      IF (opt_hstart > opt_hmax) RETURN
      IF (opt_hstart == 0.0_dp) RETURN
   END IF
   ifail = -4                ! check tolerance
   IF (tol <= 0.0_dp) RETURN
   opt_glo_tol = tol         ! default for global tolerance is tol
   IF (PRESENT(global_tol)) opt_glo_tol = global_tol
   IF (opt_glo_tol <= 0.0_dp) RETURN
   opt_loc_tol = tol         ! default for local tolerance is tol
   IF (PRESENT(local_tol)) opt_loc_tol = local_tol
   IF (opt_loc_tol <= 0.0_dp) RETURN
   opt_lin = .false.         ! default for linear is .false.
   IF (PRESENT(linear)) opt_lin = linear
   opt_New_tol = opt_glo_tol/10.0_dp ! default for Newton_tol = tol/10.0
   IF ((.not. opt_lin) .AND. (a <= 0.0_dp)) THEN
      ifail = -5             ! check Newton-tolerance
      IF (PRESENT(Newton_tol)) opt_New_tol = Newton_tol
      IF (opt_New_tol <= 0.0_dp) RETURN
      IF (opt_New_tol > tol) RETURN ! should be at least as exact
      IF (opt_New_tol > opt_glo_tol) RETURN
   END IF
   ifail = -6                ! check method
   opt_meth = 4              ! default method is method 4
   IF (PRESENT(method)) opt_meth = method
   IF ((opt_meth /= 4) .AND. (opt_meth /= 2)) RETURN
   ifail = -8                ! check switch point
   opt_switch = -1.0_dp      ! default switch point is auto detect
   IF (PRESENT(switch)) opt_switch = switch
   IF (opt_switch >= 0.0_dp) THEN ! check user input
      IF (opt_switch < a) RETURN
      IF (opt_switch > b) RETURN
   END IF

   ifail = 0
   IF (opt_hstart <= -1.0_dp) THEN ! auto detect starting step size
      CALL ivp_starth(fcn, a, y, b, opt_hstart, opt_glo_tol, &
                      opt_New_tol, opt_lin, opt_meth, ifail, dfcn)
      IF (opt_hstart < opt_hmin) opt_hstart = opt_hmin ! readjust h_start
      IF (opt_hstart > opt_hmax) opt_hstart = opt_hmax ! to match h_min
      IF (opt_hstart == 0.0_dp) opt_hstart = TINY(1.0_dp) ! and h_max
      IF (PRESENT(h_start)) h_start = opt_hstart
      IF (ifail /= 0) RETURN ! an error occurred in ivp-starth
   END IF

   IF (opt_switch <= -1.0_dp) THEN ! auto detect switch point
      CALL ivp_switch(opt_hstart, opt_meth, opt_switch, ifail)
      IF (opt_switch < a) opt_switch = a ! readjust switch point to match
      IF (opt_switch > b) opt_switch = b ! interval [a,b]
      IF (PRESENT(switch)) switch = opt_switch ! return switch point
      IF (ifail /= 0) RETURN ! an error occurred in ivp-switch
   END IF

! ------------------------------------------------------------------------
! do integration
! ------------------------------------------------------------------------
   IF (opt_switch > a) THEN  ! Zadunaisky integration on [a,switch point]
      CALL ivp_zadunaisky(fcn, a, y, opt_switch, opt_hstart, opt_hmin, &
                          opt_glo_tol, opt_New_tol, opt_lin, opt_meth, &
                          ifail, dfcn) ! solution at x=switch is y(:)
      IF (PRESENT(h_start)) h_start = opt_hstart
      IF (PRESENT(global_tol)) global_tol = opt_glo_tol
      IF (ifail /= 0) RETURN ! an error occurred in ivp-zadunaisky
      a = opt_switch         ! we continue from the switch point
   END IF                    ! do not use Zadunaisky's method

   IF (opt_switch < b) THEN  ! do RK-pair integration on [switch,b]
      CALL ivp_rkpair(fcn, a, y, b, opt_hstart, opt_hmin, &
                      MIN(opt_hmax,b-a), opt_loc_tol, opt_meth, ifail)
      IF (PRESENT(local_tol)) local_tol = opt_loc_tol
      IF (ifail /= 0) RETURN ! an error occurred in ivp-rkpair
   END IF
   RETURN
END SUBROUTINE ivp_sing

SUBROUTINE ludcmp(A, indx)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Replace matrix A by the LU-decomposition of a rowwise permutation of
! itself. 
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1016.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in out) :: A(:,:)
!        A ... (in) matrix to decompose (out) LU-decomposition
   INTEGER(KIND = i4b), INTENT(out) :: indx(:)
!        indx ... row permutation by partial pivoting
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: vv(SIZE(A,1))
!        vv ... implicit scaling of each row
   INTEGER(KIND = i4b) :: j, n, imax
!        j ... loop counter
!        n ... dimension of system
!        imax ... pivot row
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(indx)) ) THEN
!     error in input dimensions (not checked)
   END IF

   vv = MAXVAL(ABS(A), dim=2)
   IF (ANY(vv == 0.0_dp)) THEN
      WRITE(*,*) ' singular matrix in LUDCMP '
      STOP
   END IF
   vv = 1.0_dp/vv
   DO j = 1, n
      imax = (j-1)+imaxloc(vv(j:n)*ABS(A(j:n,j))) ! find pivot row
      IF (j /= imax) THEN
         CALL swap(A(imax,:), A(j,:))
         vv(imax) = vv(j)
      END IF
      indx(j) = imax
      IF (A(j,j) == 0.0_dp) A(j,j) = TINY(1.0_dp)
      A(j+1:n,j) = A(j+1:n,j) / A(j,j) ! divide by pivot element
      A(j+1:n,j+1:n) = A(j+1:n,j+1:n) - outerprod(A(j+1:n,j), A(j,j+1:n))
   END DO
   RETURN
! ------------------------------------------------------------------------
! Local subroutines
! ------------------------------------------------------------------------
   CONTAINS

FUNCTION imaxloc(v)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Get location of largest element of vector.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.993.
! --- Arguments ---
   IMPLICIT NONE
   INTEGER(KIND = i4b) :: imaxloc
   REAL(KIND = dp), INTENT(in) :: v(:)
!        v ... vector to find location
! --- Local variables ---
   INTEGER(KIND = i4b) :: imax(1)
!        imax ... vector of length 1 to use MAXLOC
! --- Code ---
   imax = MAXLOC(v(:))
   imaxloc = imax(1)
   RETURN
END FUNCTION imaxloc

SUBROUTINE swap(v1, v2)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Swap two vectors.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.991.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in out) :: v1(:), v2(:)
!        v1, v2 ... vectors to swap
! --- Local variables ---
   REAL(KIND = dp) :: dum(SIZE(v1))
!        dum ... auxiliary data
! --- Code ---
   dum = v1
   v1 = v2
   v2 = dum
   RETURN
END SUBROUTINE swap

FUNCTION outerprod(v1, v2)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Calculate the matrix of the outer product of two vectors, optimised for
! parallel processing.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1000.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: v1(:), v2(:)
!        v1, v2 ... vectors to calculate outer product
   REAL(KIND = dp) :: outerprod(SIZE(v1),SIZE(v2))
! --- Code ---
   outerprod = SPREAD(v1, dim=2, ncopies=SIZE(v2)) * &
               SPREAD(v2, dim=1, ncopies=SIZE(v1))
   RETURN
END FUNCTION outerprod
END SUBROUTINE ludcmp

SUBROUTINE lubksb(A, indx, b)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Solve a linear equation system Ax=b with the LU-decomposition and
! permutation vector at hand (to be called after LUDCMP).
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1017.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:)
!        A ... LU-decomposition
   INTEGER(KIND = i4b), INTENT(in) :: indx(:)
!        indx ... row permutation by partial pivoting
   REAL(KIND = dp), INTENT(in out) :: b(:)
!        b ... (in) right side of LES, (out) solution of LES
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: summ
!        summ ... auxiliary data
   INTEGER(KIND = i4b) :: i, n, ii, ll
!        i ... loop counter
!        n ... dimension of system
!        ii, ll ... auxiliary indices
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(indx)) .OR. (n /= SIZE(b)) ) THEN
!     error in input dimensions (not checked)
   END IF

   ii = 0
   DO i = 1, n
      ll = indx(i)
      summ = b(ll)
      b(ll) = b(i)
      IF (ii /= 0) THEN
         summ = summ - DOT_PRODUCT(A(i,ii:i-1), b(ii:i-1))
      ELSE IF (summ /= 0.0_dp) THEN
         ii = i
      END IF
      b(i) = summ
   END DO
   DO i = n,1,-1
      b(i) = (b(i) - DOT_PRODUCT(A(i,i+1:n), b(i+1:n)))/A(i,i)
   END DO
   RETURN
END SUBROUTINE lubksb

SUBROUTINE ivp_starth(fcn, a, y, b, h_guessed, tol, Newton_tol, linear, &
                      method, ifail, dfcn)
! Version 1.01, last modified 27-01-1998.
! --- Function and Algorithm ---
! Guess an initial step size that yields global error below tol after
! integration with specified method.
! --- Bibliography ---
! E.Hairer, S.P.Norsett, G.Wanner,
! 'Solving Ordinary Differential Equations I', Springer (1987), p.182.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t,y)      ! right side of the ivp, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
      FUNCTION dfcn(t,y)     ! Jacobi-matrix of fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: dfcn(SIZE(y), SIZE(y))
      END FUNCTION dfcn
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: a, b, tol, Newton_tol, y(:)
!        a, b ... interval to perform integration from a to b
!        y ... initial value y(a)
!        tol ... tolerance
!        Newton_tol ... tolerance of Newton-iteration for implicit Euler
   REAL (KIND = dp), INTENT(out) :: h_guessed
!        h_guessed ... guessed step size
   LOGICAL, INTENT(in) :: linear
!        linear ... the initial value problem is linear
   INTEGER(KIND = i4b), INTENT(in) :: method
!        method ... order of method
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!                   0 = everything is oK
!                  -1 = illegal interval range
!                  -4 = illegal tolerance
!                  -5 = illegal Newton-tolerance
!                  -6 = illegal method
!                   2 = Newton-iteration needed too many steps
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: w1(SIZE(y)), w2(SIZE(y)), h1, h2, normf, den
!        w1,w2 ... two function evaluations
!        h1,h2 ... two estimated step sizes
!        normf ... norm of function evaluation
!        den ... effect of interval range
! ------------------------------------------------------------------------
! Called software components
! ------------------------------------------------------------------------
!        indirect ... ludcmp, lubksb
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
! check input and set up variables
   ifail = -1                ! check interval
   IF (a < 0.0_dp) RETURN
   IF (b <= a) RETURN
   ifail = -4                ! check tolerance
   IF (tol <= 0.0_dp) RETURN
   IF (.not. linear) THEN
      ifail = -5             ! check Newton-tolerance
      IF (Newton_tol <= 0.0_dp) RETURN
      IF (Newton_tol > tol) RETURN ! should be at least as exact
   END IF
   ifail = -6                ! check method
   IF ((method /= 2) .AND. (method /= 4)) RETURN
   ifail = 0

   IF (b < 1.0_dp ) THEN
      den = 1.0_dp
   ELSE
      den = 1.0_dp/MAX(ABS(a),ABS(b))**method
   END IF
   IF (a > 0.0_dp) THEN      ! classic method
      w1 = fcn(a,y)          ! function evaluation at initial point
      normf = MAXVAL(ABS(w1))**method
      h1 = (tol/(den+normf))**(1.0_dp/method)
      h1 = MIN(h1, ABS(b-a))
      w2 = y + h1*w1         ! Euler step with proposed h1
   ELSE                      ! variation for singular ivp's
      normf = 0.0_dp         ! this is exact for transformed systems of
                             ! second order without positive eigenvalues
      h1 = (tol/(den+normf))**(1.0_dp/method)
      h1 = MIN(h1, ABS(b-a))
      IF (linear) THEN       ! implicit Euler step with proposed h1
         w2 = impl_lin_euler(h1)
      ELSE
         w2 = impl_nlin_euler(h1)
      END IF
   END IF
   w2 = fcn(a+h1, w2)        ! function evaluation at new initial point
   normf = MAXVAL(ABS(w2))**method
   h2 = (tol/(den+normf))**(1.0_dp/method)
   h_guessed = MIN(h1,h2)    ! choose smaller of the two step sizes
   RETURN
! ------------------------------------------------------------------------
! Local subroutines
! ------------------------------------------------------------------------
   CONTAINS

FUNCTION impl_lin_euler(h)
! Version 1.00, last modified 27-01-1998.
! --- Function and Algorithm ---
! Perform an implicit Euler step for linear fcn.
! required temporary storage: SIZE(y)^2+3*SIZE(y)+2
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_lin_euler(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
! --- Global variables ---
!        fcn, dfcn, a, y
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y))
!        Jac ... Jacobi-matrix for given x
!        w ... inhomogeous part of fcn
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
! --- Called software components ---
!        ludcmp, lubksb
! --- Code ---
! set up linear system Jac*y1=w for implicit Euler
   Jac = -h*dfcn(a + h, y)   ! homogenous component of problem
   w = 0.0_dp
   w = y + h*fcn(a + h,w)    ! inhomogenous component of problem
   DO i = 1, SIZE(y)         ! loop over diagonal, add identity
      Jac(i,i) = Jac(i,i) + 1.0_dp
   END DO ! DO i over diagonal

! solve linear system A*y1=w, destroying A and yielding y1 in w
   CALL ludcmp(Jac, indx)
   CALL lubksb(Jac, indx, w)

   impl_lin_euler = w        ! y1 = new solution for y
   RETURN
END FUNCTION impl_lin_euler

FUNCTION impl_nlin_euler(h)
! Version 1.00, last modified 27-01-1998.
! --- Function and Algorithm ---
! Perform an implicit Euler step for non-linear fcn, using
! Newton-iteration.
! required temporary storage: SIZE(y)^2+4*SIZE(y)+6
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_nlin_euler(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
! --- Global variables ---
!        fcn, dfcn, a, y, Newton_tol and constant max_newton_steps
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), nrmF, &
                      nrmdeltak, y1(SIZE(y))
!        Jac ... Jacobi-matrix for given x
!        w ... fcn call
!        nrmF ... norm of residual
!        nrmdeltak ... norm of Newton-increment (deltak)
!        y1 ... approximation of step
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i, index
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
!        index ... iteration count in Newton-iteration
! --- Called software components ---
!        ludcmp, lubksb
! --- Code ---
! get initial approximation for y1
   y1 = y

! start Newton-iteration
   index = 0
   DO
      index = index + 1

! set up linear system Jac*y1^(i) = w for Newton-increment
      w = y+h*fcn(a + h, y1) - y1 ! calculate Newton-residual and size
      nrmf = MAXVAL(ABS(w))  
      Jac = h*dfcn(a + h, y1) ! Jacobi-matrix
      DO i = 1, SIZE(y)      ! loop over diagonal, subtract identity
         Jac(i,i) = Jac(i,i) - 1.0_dp
      END DO ! DO i over diagonal

! solve linear system, destroying A and yielding delta y1 in w
      CALL ludcmp(Jac, indx)
      CALL lubksb(Jac, indx, w)

! iterate solution and exit if accurate
      y1 = y1 - w
      nrmdeltak = MAXVAL(ABS(w)) ! size of Newton-increment
      IF ((nrmf < Newton_tol) .AND. (nrmdeltak < Newton_tol)) EXIT

      IF (index > max_newton_steps) THEN
         ifail = 2           ! too many iterations
         EXIT
      END IF
   END DO ! Newton-loop

   impl_nlin_euler = y1      ! iterated y1 = new solution for y
   RETURN
END FUNCTION impl_nlin_euler
END SUBROUTINE ivp_starth

SUBROUTINE ivp_switch(h, method, switch_point, ifail)
! Version 1.02, last modified 23-02-1998.
! --- Function and Algorithm ---
! Guess a good point for specified method to switch from global error
! control to local error control.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: h
!        h ... used step size
   INTEGER(KIND = i4b), INTENT(in) :: method
!        method ... choose switch for a particular method
   REAL(KIND = dp), INTENT(out) :: switch_point
!        switch_point ... point from where local error control is used
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, not used (for future expansion)
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   ifail = 0
   IF (method == 4) THEN
      switch_point = h * switch_count_4
   ELSE IF (method == 2) THEN
      switch_point = h * switch_count_2
   END IF
   RETURN
END SUBROUTINE ivp_switch

SUBROUTINE ivp_zadunaisky(fcn, a, y, b, h, h_min, global_tol, &
                          Newton_tol, linear, method, ifail, dfcn)
! Version 1.06, last modified 23-02-1998.
! --- Function and Algorithm ---
! Do the integration of y'(t) = fcn(t, y) from a to b with constant step
! size and check the global error at b with Zadunaisky's technique against
! the global tolerance, readjust h and repeat integration if necessary.
! --- Bibliography ---
! P.E.Zadunaisky,
! 'On the Estimation of Errors Propagated in the Numerical Integration of
! Ordinary Differential Equations', Numer.Math. 27(1976), pp. 21--39.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   USE fun, ONLY : true, truelocal ! to generate log_files
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t, y)     ! right side of the ivp, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
      FUNCTION dfcn(t, y)    ! Jacobi-matrix of fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: dfcn(SIZE(y), SIZE(y))
      END FUNCTION dfcn
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: a, b, h_min, Newton_tol
!        a, b ... interval to perform integration from a to b
!        h_min ... smallest allowed step size, may be 0.0
!        Newton_tol ... tolerance of Newton-iteration for implicit RK
   REAL(KIND = dp), INTENT(in out) :: y(:), h, global_tol
!        y ... (in) initial value y(a), (out) approximation for y(b)
!        h ... (in) starting step size, (out) last used step size
!        global_tol ... (in) tolerance, (out) estimation of global error
   LOGICAL, INTENT(in) :: linear
!        linear ... the initial value problem is linear
   INTEGER(KIND = i4b), INTENT(in) :: method
!        method ... to choose a particular method out of the following 
!                   2 = 1st step implicit order 3, then explicit order 2
!                   4 = 1st step implicit order 5, then explicit order 4
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!                   0 = everything is oK
!                  -1 = illegal interval range
!                  -2 = illegal smallest step size
!                  -3 = illegal starting step size
!                  -4 = illegal global tolerance
!                  -5 = illegal Newton-tolerance
!                  -6 = illegal method
!                   1 = could not complete integration due to h < h_min
!                   2 = Newton-iteration needed too many steps
!                   3 = global error did not decrease during iteration
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   INTEGER(KIND = i4b) :: j, iter, step, pi, feldi, se, si, p, pord, &
                          maxsteps
!        j ... loop counter
!        iter ... iteration counter until global error is below tolerance
!        step ... step counter in integration from a to b
!        pi ... step counter for little steps for Zadunaisky-polynomials
!        feldi ... reference value for interpolation in yfield
!        se ... number of stages of explicit Runge-Kutta method
!        p  ... order of explicit Runge-Kutta method
!        pord ... order of Zadunaisky-polynomials
!        maxsteps ... total number of steps in the integration
   REAL(KIND = dp), ALLOCATABLE :: ae(:,:), be(:), ce(:), &
                                   ai(:,:), bi(:), ci(:)
!        ae, be, ce ... coefficients of Butcher array explicit Runge-Kutta
!        ai, bi, ci ... coefficients of Butcher array implicit Runge-Kutta
   REAL(KIND = dp) :: ya(SIZE(y)), x, h_used, global_est, ratio, p_est, &
                      old_h, old_est
!        ya ... store initial value for several integrations
!        x ... current position during integration in [a,b]
!        h_used ... used step size
!        global_est ... estimation of global error due Zadunaisky's method
!        ratio ... step size-update ratio
!        p_est ... estimation of order of the global error
!        old_h ... last used step size
!        old_est ... estimated error of solution of last iteration
   REAL(KIND = dp), ALLOCATABLE :: yfield(:,:)
!        yfield ... store all approximations to interpolate the solution
   CHARACTER(LEN = 12) :: name1 = 'logx.abs', & ! [x, global-err],
                          name2 = 'logx.grd', & ! [x, 0], grid
                          name3 = 'logx.est', & ! [x, global-est]
                          name4 = 'logx.gr2', & ! [x, global-err], grid
                          name5 = 'logx.lok', & ! [x, local-err]
                          name6 = 'logx.xhp'    ! [x, h]
!        name? ... names of log_files
! ------------------------------------------------------------------------
! Called software components
! ------------------------------------------------------------------------
!        indirect ... ludcmp, lubksb
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
! check input and set up variables
   ifail = -1                ! check interval
   IF (a < 0.0_dp) RETURN
   IF (b <= a) RETURN
   ifail = -2                ! check smallest step size
   IF (h_min < 0.0_dp) RETURN
   IF (h_min > b-a) RETURN
   ifail = -3                ! check starting step size
   IF (h < h_min) h = h_min
   IF (h > b-a) h = b-a
   IF (h == 0.0_dp) RETURN
   ifail = -4                ! check global tolerance
   IF (global_tol <= 0.0_dp) RETURN
   IF (.not. linear) THEN
      ifail = -5             ! check Newton-tolerance
      IF (Newton_tol <= 0.0_dp) RETURN
      IF (Newton_tol > global_tol) RETURN ! should be at least as exact
   END IF
   ifail = -6                ! check method

! set up Runge-Kutta arrays (both explicit and implicit) for used method
   p = method
   IF (method == 4) THEN
      se = 4                 ! 4 stages-method of order 4
      ALLOCATE(ae(2:se, 1:se-1), be(se), ce(2:se))
      CALL coeff_erk_the4(ae, be, ce)

      IF (a == 0.0_dp) THEN  ! use implicit method for 1st step
         si = 3              ! 3 stages-method of order 5
         ALLOCATE(ai(si,si), bi(si), ci(si))
         CALL coeff_irk_r2a5(ai, bi, ci)
      END IF
   ELSE IF (method == 2) THEN
      se = 2                 ! 2 stages-method of order 2
      ALLOCATE(ae(2:se, 1:se-1), be(se), ce(2:se))
      CALL coeff_erk_heun2(ae, be, ce)

      IF (a == 0.0_dp) THEN
         si = 2              ! 2 stages-method of order 3
         ALLOCATE(ai(si,si), bi(si), ci(si))
         CALL coeff_irk_r2a3(ai, bi, ci)
      END IF
   ELSE
      RETURN
   END IF
   pord = p+1                ! order of Zadunaisky's polynomials

   ifail = 0
   ya = y                    ! store initial value y(a)
   iter = 0
   p_est = p
   old_h = h
   old_est = HUGE(1.0_dp)
! ------------------------------------------------------------------------
! do iteration until global error is below tolerance
! ------------------------------------------------------------------------
   DO
      iter = iter + 1

      IF (log_file) THEN     ! open log files
         name1(4:4) = char(iter+64)
         OPEN(14, FILE = name1)
         WRITE(14, *) a, MAXVAL(ABS(ya - True(a)))
         name2(4:4) = char(iter+64)
         OPEN(20, FILE = name2)
         WRITE(20, *) a, 0.0_dp
         name4(4:4) = char(iter+64)
         OPEN(21, FILE = name4)
         WRITE(21, *) a, MAXVAL(ABS(ya - True(a)))
         name5(4:4) = char(iter+64)
         OPEN(22, FILE = name5)
         WRITE(22, *) a, 0.0_dp
         name6(4:4) = char(iter+64)
         OPEN(23, FILE = name6)
         WRITE(23, *) a, 0.0_dp
      END IF

! calculate amount of necessary steps and adjust h to distribute grid
! points equally spaced on [a,b]
      maxsteps = CEILING((b-a)/h/(1+EPSILON(1.0_dp)))
                             ! use h*(1+EPSILON) to avoid one further step
                             ! due to roundoff error in h
      h = (b-a)/maxsteps     ! hnew <= h
      h_used = h/pord        ! used step size for little steps
!     WRITE(*,*) ' Maxsteps: ', maxsteps
!     WRITE(*,*) ' Step size: ',h

! check step size again
      IF ((h < h_min) .OR. (h_used <= b*EPSILON(1.0_dp))) THEN
         IF (iter == 1) THEN
            ifail = -3       ! initial step size too small
         ELSE                ! iter > 1, so there is a solution available
            ifail = 1        ! could not complete because of h < h_min
            global_tol = old_est ! that is what we got in last iteration
            h = old_h        ! with the last used step size
         END IF
         EXIT
      END IF

      ALLOCATE(yfield(0:maxsteps*pord, SIZE(y)))

! calculate first approximation
      y = ya                 ! initialise initial values
      x = a
      step = 0
      yfield(step,:) = y     ! store first value, the initial value
      DO j = 1, maxsteps     ! loop for h-steps
         DO pi = 1, pord     ! loop for little steps
            IF ((j == 1) .AND. & ! use implicit method for first step
               (pi == 1) .AND. (a == 0.0_dp)) THEN
               IF (linear) THEN
                  y = y + h_used*impl_lin_step(h_used,.false.)
               ELSE
                  y = y + h_used*impl_nlin_step(h_used,.false.)
                  IF (ifail == 2) THEN ! stop integration due to error
                     DEALLOCATE(yfield) ! freeing temporary data
                     EXIT
                  END IF
               END IF
            ELSE
               y = y + h_used*expl_step(h_used,.false.)
            END IF
            x = x + h_used   ! go to next grid point
            step = step + 1
            yfield(step,:) = y ! store approximation
            IF (log_file) WRITE(14, *) x, MAXVAL(ABS(y - True(x)))
            IF (log_file) WRITE(22, *) x, MAXVAL(ABS(y - &
                                TrueLocal(x, x-h_used, yfield(step-1,:))))
         END DO ! DO pi over little steps
         IF (log_file) THEN
            WRITE(20, *) x, 0.0_dp
            WRITE(21, *) x, MAXVAL(ABS(y - True(x)))
            WRITE(23, *) x, h
         END IF
      END DO ! DO j over steps

      IF (log_file) THEN
         CLOSE(14)
         CLOSE(20)
         CLOSE(21)
         CLOSE(22)
         CLOSE(23)
         name3(4:4) = char(iter+64)
         OPEN(15, FILE = name3)
         WRITE(15,*) a, 0.0_dp
      END IF

! calculate approximation for the neighbouring problem
      x = a                  ! initialise initial values
      step = 0
      y = yfield(step,:)
      DO j = 1, maxsteps     ! loop for h-steps
         feldi = step
         DO pi = 1, pord     ! loop for little steps
            IF ((j == 1) .AND. & ! use implicit method for first step
               (pi == 1) .AND. (a == 0.0_dp)) THEN
               IF (linear) THEN
                  y = y + h_used*impl_lin_step(h_used,.true.)
               ELSE
                  y = y + h_used*impl_nlin_step(h_used,.true.)
                  IF (ifail == 2) THEN ! stop integration due to error
                     DEALLOCATE(yfield) ! freeing temporary data
                     EXIT
                  END IF
               END IF
            ELSE
               y = y + h_used*expl_step(h_used,.true.)
            END IF
            x = x + h_used   ! go to next grid-points
            step = step +1
            IF (log_file) WRITE(15, *) x, MAXVAL(ABS(y - yfield(step,:)))
         END DO ! DO pi over little steps
      END DO ! DO j over steps

! estimate global error with solution of the last step
      global_est = MAXVAL(ABS(y - yfield(step,:)))
      IF (log_file) CLOSE(15)
!     WRITE(*,*) ' delta ', global_est
      y = yfield(step,:)     ! store approximation before
      DEALLOCATE(yfield)     ! freeing temporary data

      IF (global_est < global_tol) THEN ! accept solution
         global_tol = global_est
         ifail = 0           ! oK, we did it
         EXIT
      ELSE                   ! reject solution, calculate new step size
         IF (iter > 1) THEN  ! after second iteration, estimate order
            IF ( (old_est > 0.0_dp) .AND. (global_est > 0.0_dp) ) THEN
               p_est = LOG(old_est/global_est)/LOG(old_h/h)
               IF ((p_est < 2.0_dp) .OR. (p_est > p)) p_est = p
!              WRITE(*,*) ' order estimation ', p_est
            END IF
         END IF
         old_h = h           ! store old h and error to estimate order
         IF (global_est >= old_est) THEN
            ifail = 3        ! error is not decreasing
            global_tol = global_est ! we got with the last step size
            EXIT
         END IF
         old_est = global_est

         ratio = ((global_est/global_tol)**(1.0_dp/p_est))
         ratio = ratio/safety_factor_zadu
         h = h/ratio         ! reduced step size by update ratio
!        WRITE(*,*) ' update ratio ', ratio
!        WRITE(*,*) ' new step size ', h
      END IF
   END DO ! iter

! restore workspace
   IF (a == 0.0_dp) DEALLOCATE(ci, bi, ai) ! deallocate in reversed order
   DEALLOCATE(ce, be, ae)
   RETURN
! ------------------------------------------------------------------------
! Local subroutines
! ------------------------------------------------------------------------
   CONTAINS

SUBROUTINE coeff_erk_heun2(a, b, c)
! Version 1.01, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the explicit Butcher array with 'Heun's Method'.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(2:2, 1:1), b(2), c(2:2)
!        a, b, c ... coefficients of Butcher array
! --- Code ---
   a(2,1) = 1.0_dp
   b = (/ 1.0_dp/2, 1.0_dp/2 /)
   c = a(2,1)                ! apply row-condition
   RETURN
END SUBROUTINE coeff_erk_heun2

SUBROUTINE coeff_irk_r2a3(a, b, c)
! Version 1.01, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the implicit Butcher array with 'Radau IIA order 3'.
! --- Bibliography ---
! E.Hairer, G.Wanner,
! 'Solving Ordinary Differential Equations II', Springer (1991), p.74.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(2,2), b(2), c(2)
!        a, b, c ... coefficients of Butcher array
! --- Code ---
   a(1,:) = (/ 5.0_dp/12, -1.0_dp/12 /)
   a(2,:) = (/ 3.0_dp/4, 1.0_dp/4 /)
   b = a(2,:)
   c = (/ 1.0_dp/3, 1.0_dp /)
   RETURN
END SUBROUTINE coeff_irk_r2a3

SUBROUTINE coeff_erk_the4(a,b,c)
! Version 1.01, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the explicit Butcher array with the classic Runge-Kutta.
! --- Bibliography ---
! E.Hairer, S.P.Norsett, G.Wanner,
! 'Solving Ordinary Differential Equations I', Springer (1987), p.137.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(2:4, 1:3), b(4), c(2:4)
!        a, b, c ... coefficients of Butcher array
! --- Code ---
   a = 0.0_dp
   a(2,1) = 1.0_dp/2
   a(3,2) = 1.0_dp/2
   a(4,3) = 1.0_dp
   b = (/ 1.0_dp/6, 2.0_dp/6, 2.0_dp/6, 1.0_dp/6 /)
   c = (/ 1.0_dp/2, 1.0_dp/2, 1.0_dp /)
   RETURN
END SUBROUTINE coeff_erk_the4

SUBROUTINE coeff_irk_r2a5(a,b,c)
! Version 1.02, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the implicit Butcher array with 'Ehle/Radau IIA'.
! --- Bibliography ---
! E.Hairer, G.Wanner,
! 'Solving Ordinary Differential Equations II', Springer (1991), p.74.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(3,3), b(3), c(3)
!        a, b, c ... coefficients of Butcher array
! --- Local variables ---
   REAL(KIND = dp) :: sq6
!        sq6 ... store SQRT(6.0)
! --- Code ---
   sq6 = SQRT(6.0_dp)
   a(1,:) = (/ (88.0_dp-7.0_dp*sq6)/360, (296.0_dp-169.0_dp*sq6)/1800, &
               (-2.0_dp+3.0_dp*sq6)/225 /)
   a(2,:) = (/ (296.0_dp+169.0_dp*sq6)/1800, (88.0_dp+7.0_dp*sq6)/360, &
               (-2.0_dp-3.0_dp*sq6)/225 /)
   a(3,:) = (/ (16.0_dp-sq6)/36, (16.0_dp+sq6)/36, 1.0_dp/9 /)
   b = a(3,:)
   c = (/ (4.0_dp-sq6)/10, (4.0_dp+sq6)/10, 1.0_dp /)
   RETURN
END SUBROUTINE coeff_irk_r2a5

FUNCTION lagrang(x)
! Version 1.00, last modified 14-01-1998.
! --- Function and Algorithm ---
! Evaluate the polynomial sum_{j=0}^pord omega_j*f_j (lagrange-polynomial)
! interpolating yfield on [a,a+pord*h] at x.
! required temporary storage: 2*SIZE(y)+5
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: lagrang(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: x
!        x ... posititon the evaluation of lagrange-polynomial is required
! --- Global variables ---
!        yfield(:,:), feldi, pord, h_used, y (size only)
! --- Local variables ---
   REAL(KIND = dp) :: u, w, ph(SIZE(y))
!        u ... x-value shifted to [0, pord*h]
!        w ... auxiliary data
!        ph ... value of lagrange-polynomial at x
   INTEGER(KIND = i4b) :: j, k
!        j, k ... loop counters
! --- Code ---
   u = x - a - h_used*feldi  ! move interval [a+...,...] to [0,...]
   ph = 0.0_dp               ! initialise sum of omega_j*f_j
   DO j = 0, pord            ! calculate sum omega_j*f_j (pord+1 summands)
      w = 1.0_dp
      DO k = 0, pord         ! calculate product omega_j (pord factors)
         IF (j /= k) w = w * (u-h_used*k)/(j-k)
      END DO
      ph = ph + w*yfield(feldi+j,:) ! add omega_j*f_j
   END DO
   lagrang = ph/h_used**pord
   RETURN
END FUNCTION lagrang

FUNCTION difflagr(x)
! Version 1.00, last modified 15-01-1998.
! --- Function and Algorithm ---
! Evaluate the derivative of the polynomial sum_{j=0}^pord omega_j*f_j
! (lagrange-polynomial) interpolating yfield on [a,a+pord*h] at x.
! required temporary storage: 2*SIZE(y)+7
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: difflagr(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: x
!        x ... value the evaluation is required
! --- Global variables ---
!        yfield(:,:), feldi, pord, h_used, y (size only)
! --- Local variables ---
   REAL(KIND = dp) :: u, w, w2, phs(SIZE(y))
!        u ... x-value shifted to [0,pord*h]
!        w, w2 ... auxiliary data
!        phs ... value of (lagrange-polynomial)' at x
   INTEGER(KIND = i4b) :: j, k, l
!        j, k, l ... loop counter
! --- Code ---
   u = x - a - h_used*feldi  ! move interval [a+...,...] to [0,...]
   phs = 0.0_dp              ! initialise sum of omega'_j*f_j

   DO j = 0, pord            ! calculate sum omega'_j*f_j(pord+1 summands)
      w = 0.0_dp             ! initialise sum for omega'_j
      DO k = 0, pord         ! calculate sum omega'_j (pord summands)
         IF (j /= k) THEN
            w2 = 1.0_dp      ! initialise k-th product in omega'_j
            DO l = 0, pord   ! calculate k-th product in omega'_j
               IF ((j /= l).and.(l /= k)) w2 = w2 * (u-h_used*l)
            END DO
            w = w + w2       ! add components of omega'_j
         END IF
      END DO

      DO k = 0, pord         ! calculate the constant factor of omega_j
         IF (j /= k ) w = w/(j-k)
      END DO
      phs = phs + w*yfield(feldi+j,:) ! add omega'_j*f_j
   END DO
   difflagr = phs/h_used**pord
   RETURN
END FUNCTION difflagr

FUNCTION expl_step(h, zadu)
! Version 1.03, last modified 16-02-1998.
! --- Function and Algorithm ---
! Perform one explicit Runge-Kutta step for fcn or neighbouring problem.
! required temporary storage: (se+3)*SIZE(y)+4
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: expl_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform explicit Runge-Kutta step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, x, y, Butcher array(ae, be, ce, se), y (size only)
!        for zadu indirect usage of yfield(:,:), feldi, pord, h_used
! --- Local variables ---
   REAL(KIND = dp) :: w(SIZE(y)), f(se, SIZE(y)), u, lag(SIZE(y))
!        w ... subsequent stages of Runge-Kutta algorithm
!        u ... auxiliary data
!        f ... store fcn calls
!        lag ... explicit evaluation of lagrang for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: i
!        i ... loop counter
! --- Called software components ---
!        for zadu direct call of lagrang, difflagr
! --- Code ---
   f(1, :) = Fcn(x, y)       ! first function evaluation
   IF (zadu) THEN 
      lag = lagrang(x)
      f(1, :) = f(1, :) + difflagr(x) - Fcn(x, lag)
   END IF
   DO i = 2, se              ! loop over stages
      u = x + ce(i)*h
      w = y + h*MATMUL(ae(i, 1:i-1), f(1:i-1, :))
      f(i, :) = Fcn(u, w)
      IF (zadu) THEN
         lag = lagrang(u)
         f(i, :) = f(i, :) + difflagr(u) - Fcn(u, lag)
      END IF
   END DO ! DO i over stages
   expl_step = MATMUL(be, f) ! this is increment for new solution for y
   RETURN
END FUNCTION expl_step

FUNCTION impl_lin_step(h, zadu)
! Version 1.02, last modified 16-02-1998.
! --- Function and Algorithm ---
! Perform one implicit step for linear fcn or neighbouring problem.
! required temporary storage: (si^2+si+1)*SIZE(y)^2+(2*si+3)*SIZE(y)+8
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_lin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Runge-Kutta step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, x, y, Butcher array(ai, bi, ci, si), y (size only)
!        for zadu indirect usage of yfield(:,:), feldi, pord, h_used
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), &
                      A(si*SIZE(y),si*SIZE(y)), b(si*SIZE(y)), &
                      AM(SIZE(y),si*SIZE(y)), bm(SIZE(y)), u, lag(SIZE(y))
!        Jac ... auxiliary data to get Jacobi-matrix of fcn for given x
!        w ... auxiliary data to get inhomogenous part of fcn
!        A, b ... linear system for the implicit coefficients k
!        AM, bm ... homogenous and inhomogenous part stored
!        u ... auxiliary data
!        lag ... explicit evaluation of lagrang for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: indx(si*SIZE(y)), i, i1, i2, j, j1, j2
!        indx ... pivoting indices for LU-decomposition
!        i, j ... loop counter
!        i1, i2, j1, j2 ... precalculated indices
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu lagrang, difflagr
! --- Code ---
! set up linear system A*k=b for coefficients of the implicit step
   b = 0.0_dp
   bm = 0.0_dp
   DO i = 1, si              ! loop over stages
      i1 = (i-1)*SIZE(y)+1   ! precalculate indices
      i2 = i*SIZE(y)
      w = 0.0_dp
      u = x + ci(i)*h
      w = fcn(u,w)           ! inhomogenous component of problem
      IF (zadu) THEN
         lag = lagrang(u)
         w = w + difflagr(u) - fcn(u, lag)
      END IF
      Jac = dfcn(u,y)        ! homogenous component of problem
      bm = bm + w*bi(i)
      AM(:,i1:i2) = Jac*bi(i)
      DO j = 1, si
         j1 = (j-1)*SIZE(y)+1
         j2 = j*SIZE(y)
         b(j1:j2) = b(j1:j2) - h*w*ai(j,i)
         A(j1:j2,i1:i2) = Jac*ai(j,i)
      END DO ! DO j over stages
      b(i1:i2) = b(i1:i2) - y
   END DO ! DO i over stages
   A = A*h
   DO i = 1, si*SIZE(y)      ! loop over diagonal, subtract identity
      A(i,i) = A(i,i) - 1.0_dp
   END DO ! DO i over diagonal

! solve linear system A*k=b, destroying A and yielding k in b
   CALL ludcmp(A, indx)
   CALL lubksb(A, indx, b)

! new solution for y is found similar to MATMUL(bi, f(.,b))
   impl_lin_step = MATMUL(AM, b) + bm
   RETURN
END FUNCTION impl_lin_step

FUNCTION impl_nlin_step(h, zadu)
! Version 1.03, last modified 16-02-1998.
! --- Function and Algorithm ---
! Perform one implicit step for non-linear fcn or the neighbouring problem
! using Newton-iteration.
! required temporary storage: (si^2+1)*SIZE(y)^2+(2*si+3)*SIZE(y)+11
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_nlin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Runge-Kutta step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y, Butcher array(ai, bi, ci, si), Newton_tol,
!        y (size only) and constant max_newton_steps
!        for zadu indirect usage of yfield(:,:), feldi, pord, h_used
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), &
                      A(si*SIZE(y),si*SIZE(y)), b(si*SIZE(y)), &
                      u, k(si*SIZE(y)), nrmF, nrmdeltak, lag(SIZE(y))
!        Jac ... auxiliary data to get Jacobi-matrix for given x
!        w ... auxiliary data for fcn calls
!        A, b ... linear system for the Newton-increment deltak
!        u ... auxiliary data
!        k ... implicit coefficients
!        nrmF ... norm of residual
!        nrmdeltak ... norm of Newton-increment (norm of deltak)
!        lag ... explicit evaluation of lagrang for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: indx(si*SIZE(y)), i, i1, i2, j, j1, j2, index
!        indx ... pivoting indices for LU-decomposition
!        i, j ... loop counter
!        i1, i2, j1, j2 ... precalculated indices
!        index ... iteration count in Newton-iteration
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu lagrang, difflagr
! --- Code ---
! get initial approximations for coefficients k of implicit step
   DO i = 1, si              ! loop over stages,
      k((i-1)*SIZE(y)+1:i*SIZE(y)) = y
   END DO ! DO i over stages

! start Newton-iteration
   index = 0
   DO
      index = index + 1
      b = 0.0_dp

! set up linear system A*deltak^(i)=b for Newton-increment
      DO i = 1, si           ! loop over stages
         i1 = (i-1)*SIZE(y)+1
         i2 = i*SIZE(y)
         u = x + ci(i)*h
         w = fcn(u, k(i1:i2))
         IF (zadu) THEN
            lag = lagrang(u)
            w = w + difflagr(u) - fcn(u, lag)
         END IF
         Jac = dfcn(u, k(i1:i2)) ! Jacobi-matrix
         DO j = 1, si
            j1 = (j-1)*SIZE(y)+1
            j2 = j*SIZE(y)
            b(j1:j2) = b(j1:j2) - h*w*ai(j,i)
            A(j1:j2,i1:i2) = Jac*ai(j,i)
         END DO ! DO j over stages
         b(i1:i2) = b(i1:i2) - y
      END DO ! DO i over stages
      b = b + k
      A = A * h
      DO i = 1, si*SIZE(y)   ! loop over diagonal, subtract identity
         A(i,i) = A(i,i) - 1.0_dp
      END DO ! DO i over diagonal
      nrmf = MAXVAL(ABS(b))  ! Newton-residual for this approximation

! solve linear system, destroying A and yielding deltak in b
      CALL ludcmp(A, indx)
      CALL lubksb(A, indx, b)

! iterate solution and exit if accurate
      k = k + b
      nrmdeltak = MAXVAL(ABS(b)) ! size of Newton-increment
      IF ((nrmf < Newton_tol) .AND. (nrmdeltak < Newton_tol)) EXIT

      IF (index > max_newton_steps) THEN
         ifail = 2           ! error, too many iterations
         impl_nlin_step = 0.0_dp
         RETURN
      END IF
   END DO ! Newton-loop

! calculate new solution for y
   b = 0.0_dp
   DO i = 1, si              ! loop over stages
      u = x + ci(i)*h
      w = fcn(u, k((i-1)*SIZE(y)+1:i*SIZE(y)))
      IF (zadu) THEN
         lag = lagrang(u)
         w = w + difflagr(u) - fcn(u, lag)
      END IF
      b(1:SIZE(y)) = b(1:SIZE(y)) + w*bi(i)
   END DO ! DO i over stages
   impl_nlin_step = b(1:SIZE(y))
   RETURN
END FUNCTION impl_nlin_step
END SUBROUTINE ivp_zadunaisky

SUBROUTINE ivp_rkpair(fcn, a, y, b, h, h_min, h_max, local_tol, method, &
                      ifail)
! Version 1.03, last modified 23-02-1998.
! --- Function and Algorithm ---
! Solve the ODE y'(t) = fcn(t, y) in [a,b] with variable step size using
! an embedded pair to check the local error against the local tolerance.
! --- Bibliography ---
! J.R.Dormand,
! 'Numerical Methods for Differential Equations', CRC Press (1996).
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   USE fun, ONLY : true, truelocal ! to generate log_files
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t, y)     ! right side of the ivp, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: b, h_min, h_max
!        b ... right bound of interval to integrate
!        h_min ... smallest allowed step size, may be 0.0
!        h_max ... maximum allowed step size
   REAL(KIND = dp), INTENT(in out) :: a, y(:), h, local_tol
!        a ... (in) left bound, (out) end position of integration
!        y ... (in) initial value y(a), (out) approximation for y(b)
!        h ... (in) starting step size, (out) last used step size
!        local_tol ... (in) tolerance, (out) estimation of local error
   INTEGER(KIND = i4b), INTENT(in) :: method
!        method ... to choose a particular method out of the following 
!                   2 = explicit 2-3 pair (Dormand)
!                   4 = explicit 4-5 pair (Runge-Kutta-Fehlberg)
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!                   0 = everything is oK
!                  -1 = illegal interval range
!                  -2 = illegal smallest step size
!                  -3 = illegal starting step size
!                  -4 = illegal local tolerance
!                  -6 = illegal method
!                  -7 = illegal maximum step size
!                   1 = could not complete integration due to h < h_min
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   LOGICAL :: last_step
!        last_step ... last step is to be performed
   INTEGER(KIND = i4b) :: i, s, p
!        i ... loop counter
!        s ... number of stages of explicit Runge-Kutta pair
!        p  ... order of lower Runge-Kutta pair method
   REAL(KIND = dp), ALLOCATABLE :: ae(:,:), be(:), ble(:), ce(:), f(:,:)
!        ae, be, ble, ce ... Butcher array of Runge-Kutta pair
!        f ... store fcn calls
   REAL(KIND = dp) :: w(SIZE(y)), x, local_est, ratio, oop, old_h, old_est
!        w ... subsequent stages of Runge-Kutta algorithm
!        x ... current position during integration in [a,b]
!        local_est ... estimation of the local error from Runge-Kutta pair
!        ratio ... step size update ratio
!        oop ... exponent in error estimation ( 1/(p+1) )
!        old_h ... last used step size
!        old_est ... estimated local error of last step
   CHARACTER(LEN = 12) :: name1 = 'log0.abs', & ! [x, global-err], grid
                          name2 = 'log0.grd', & ! [x, 0], grid
                          name3 = 'log0.est', & ! [x, local-est]
                          name5 = 'log0.lok', & ! [x, local-err]
                          name6 = 'log0.xhp'    ! [x, h]
!        name? ... names of log_files
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
! check input and set up variables
   ifail = -1                ! check interval
   IF (a < 0.0_dp) RETURN
   IF (b <= a) RETURN
   ifail = -2                ! check smallest step size
   IF (h_min < 0.0_dp) RETURN
   IF (h_min > b-a) RETURN
   ifail = -7                ! check maximum step size
   IF (h_max <= 0.0_dp) RETURN
   IF (h_max > b-a) RETURN
   IF (h_max < h_min) RETURN
   ifail = -3                ! check starting step size
   IF (h < h_min) h = h_min
   IF (h > h_max) h = h_max
   IF (h == 0.0_dp) RETURN
   ifail = -4                ! check local tolerance
   IF (local_tol <= 0.0_dp) RETURN
   ifail = -6                ! check method

! set up Runge-Kutta pairs
   p = method
   IF (method == 4) THEN
      s = 6                  ! 6 stages-method of order 4, 5
      ALLOCATE(ae(2:s, 1:s-1), be(s), ble(s), ce(2:s))
      CALL coeff_pair_engl45(ae, be, ble, ce)
   ELSE IF (method == 2) THEN
      s = 3                  ! 3 stages-pair of order 2, 3
      ALLOCATE(ae(2:s, 1:s-1), be(s), ble(s), ce(2:s))
      CALL coeff_pair23(ae, be, ble, ce)
   ELSE
      RETURN
   END IF

   ALLOCATE(f(s, SIZE(y)))
   ifail = 0
   oop = 1.0_dp/(1+p)
   x = a
   last_step = .FALSE.
   old_est = 0.0_dp
   old_h = h

   IF (log_file) THEN        ! open log files
      OPEN(14, FILE = name1)
      OPEN(15, FILE = name3)
      WRITE(15,*) a, 0.0_dp
      OPEN(20, FILE = name2)
      OPEN(22, FILE = name5)
      OPEN(23, FILE = name6)
   END IF

! ------------------------------------------------------------------------
! do integration until right bound is reached
! ------------------------------------------------------------------------
   DO                        ! loop over steps
      f(1, :) = Fcn(x, y)    ! first function evaluation
      DO i = 2, s            ! loop over stages
         w = y + h*MATMUL(ae(i, 1: i-1), f(1: i-1, :))
         f(i, :) = Fcn(x + ce(i)*h, w)
      END DO ! DO i over stages

      local_est = h*MAXVAL(ABS(MATMUL(ble - be, f))) ! local error
      ratio = ((local_est/local_tol)**oop)/safety_factor_pair

      IF (local_est < local_tol) THEN ! accepts step, update ratio < 1
         y = y + h*MATMUL(be, f) ! new solution for y with the higher
         x = x + h           ! order method

         old_est = local_est ! store the last local error estimation
         old_h = h           ! and the last used step size

         IF (log_file) THEN
            WRITE(14, *) x, MAXVAL(ABS(y - True(x)))
            WRITE(15, *) x, local_est
            WRITE(20, *) x, 0.0_dp
            WRITE(22, *) x, MAXVAL(ABS(y - &
                                TrueLocal(x, x-h, y-h*MATMUL(be,f))))
            WRITE(23, *) x, h
         END IF

         IF ( (last_step) .OR. (x >= b) ) EXIT
         h = h/MAX(ratio, 0.1_dp) ! update step
         IF (h > h_max) h = h_max
         IF(x + h > b) THEN  ! hit end point at next step
            h = b - x
            last_step = .TRUE.
         END IF
      ELSE                   ! reject step, update ratio >= 1
         h = h/MIN(ratio, 10.0_dp) ! reduced step
         IF (last_step) last_step = .FALSE. ! repeat last step
      END IF

! check step size
      IF ( (h < h_min) .OR. (h <= b*EPSILON(1.0_dp)) ) THEN
         ifail = 1           ! could not complete because of h < h_min
         h = old_h           ! last used step size was used for last step
         EXIT
      END IF
   END DO ! loop over steps
   a = x                     ! we reached this point, should be b
   local_tol = old_est

   IF (log_file) THEN
      CLOSE(14)
      CLOSE(15)
      CLOSE(20)
      CLOSE(22)
      CLOSE(23)
   END IF

! restore workspace
   DEALLOCATE(f)             ! deallocation in reversed order
   DEALLOCATE(ce, ble, be, ae)
   RETURN
! ------------------------------------------------------------------------
! Local subroutines
! ------------------------------------------------------------------------
   CONTAINS

SUBROUTINE coeff_pair23(a, b, bl, c)
! Version 1.01, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the explicit Butcher array with a RK3(2) pair.
! --- Bibliography ---
! J.R.Dormand,
! 'Numerical Methods for Differential Equations', CRC Press (1996), p.79.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(2:3, 1:2), b(3), bl(3), c(2:3)
!        a, b, bl, c ... coefficients of Butcher array
! --- Code ---
   a = 0.0_dp
   a(2,1) = 1.0_dp/2
   a(3,1:2) = (/ -1.0_dp, 2.0_dp /)
   b = (/ 1.0_dp/6, 2.0_dp/3, 1.0_dp/6 /)
   bl = (/ 1.0_dp/2, 0.0_dp, 1.0_dp/2 /)
   c = (/ 1.0_dp/2, 1.0_dp /)
   RETURN
END SUBROUTINE coeff_pair23

SUBROUTINE coeff_pair_engl45(a, b, bl, c)
! Version 1.01, last modified 10-02-1998.
! --- Function and Algorithm ---
! Initialise the explicit Butcher array with 'RK45 (England)'.
! --- Bibliography ---
! R.D.Grigorieff,
! 'Numerik gew"ohnlicher Differentialgleichungen, Band 1', (1972), p. 20.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a(2:6, 1:5), b(6), bl(6), c(2:6)
!        a, b, bl, c ... coefficients of Butcher array
! --- Code ---
   a = 0.0_dp
   a(2,1) = 1.0_dp/2
   a(3,1:2) = (/ 1.0_dp/4, 1.0_dp/4 /)
   a(4,1:3) = (/ 0.0_dp, -1.0_dp, 2.0_dp /)
   a(5,1:4) = (/ 7.0_dp/27, 10.0_dp/27, 0.0_dp, 1.0_dp/27 /)
   a(6,1:5) = (/ 28.0_dp/625, -1.0_dp/5, 546.0_dp/625, 54.0_dp/625, &
                 -378.0_dp/625 /)
   b = (/ 14.0_dp/336, 0.0_dp, 0.0_dp, 35.0_dp/336, 162.0_dp/336, &
          125.0_dp/336 /)
   bl = (/ 1.0_dp/6, 0.0_dp, 2.0_dp/3, 1.0_dp/6, 0.0_dp, 0.0_dp /)
   c = (/ 1.0_dp/2, 1.0_dp/2, 1.0_dp, 2.0_dp/3, 1.0_dp/5 /)
   RETURN
END SUBROUTINE coeff_pair_engl45
END SUBROUTINE ivp_rkpair

END MODULE solsiivp
